// $Header: p:\\graphics\\CVS/src/GraphicsTown/Drive.cpp,v 1.2 2000/11/29 18:49:30 gleicher Exp $
// GraphicsTown - CS559-2000, sample solution

#include "GrTown_PCH.H"

#include "Behaviors.H"
#include "GrObject.H"
#include "Objects.H"
#include "Roads.H"
#include <assert.h>

/***********************************************************************/
// implementation of random driving behavior
// cars randomly drive around the neighborhood, stopping at intersection
// no notion of collision avoidance
// turning is a hack

// some random utility routines...
static char* stateName(CruiseState s)
{
	switch(s) {
	case driving: return "driving";
	case intersection: return "intersection";
	case stopped: return "stopped";
	case turning: return "turning";
	default: return "huh?";
	};
}
static void reRange(int& dir)
{
	if (dir<0) dir +=4;
	dir = dir%4;
}

static bool isLegal(float f)
{
  if ( (f>-10000) && (f<10000)) return true;
  return false;
}
static bool isDir(float f)
{
  if ((f>=-1) && (f<=4)) return true;
  else return false;
}

//////////////////////////////////////////////////////////////////////////
Drive::Drive(GrObject* car, int startInt, int startDir) :
  lastIntersect(startInt), dir(startDir),
  nextIntersect(0), state(stopped),
  stateStart(0), Behavior(car), delay(1)
{
  // figure out the starting position
  locateEntry(startInt,startDir,fromX,fromZ);
  setMatrix(fromX,fromZ,dir);
}
  
int Drive::genNext(int from, int oldDir, int& nd)
{
  Intersection* cur = theIntersections(from);

  // the directions...
  int forward = cur->neighbor[oldDir];
  int ld = (oldDir+1) % 4;
  int left = cur->neighbor[ld];
  int rd = (oldDir+3) % 4;
  int right = cur->neighbor[rd];
  
  // only go backwards if no other direction to go...
  if (!forward && !left && !right) {
	printf("About Face!\n");
	nd = (oldDir+2 %2);
	return cur->neighbor[newDir];
  }

  // 3 permutations
  int ns[3], di[3];
  switch (rand()%4) {
  case 0: 
	ns[0]=forward; ns[1]=right; ns[2]=left;
	di[0]=oldDir;  di[1]=rd;    di[2]=ld;
	break;
  case 1: 
	ns[0]=forward; ns[1]=left;  ns[2]=right;
	di[0]=oldDir;  di[1]=ld;    di[2]=rd;
	break;
  case 2:
	ns[0]=right;   ns[1]=left;  ns[2]=forward;
	di[0]=rd;      di[1]=ld;    di[2]=oldDir;
	break;
  case 3:
	ns[0]=left;    ns[1]=right; ns[2]=forward;
	di[0]=ld;      di[1]=rd;    di[2]=oldDir;
	break;
  };
  for (int i=0; i<3; i++)
	if (ns[i]) {
	  newDir = di[i];
	  return ns[i];
	}
  printf("Didn't find a new direction?\n");
  return from;
}

// remember, dir=0 means the Z axis is facing down the positive X
// axis, which is a rotation of -90 degrees
void Drive::setMatrix(float x, float z, float dir)
{
  float ty = owner->transform[3][1];
  assert(isLegal(x));
  assert(isLegal(z));
  assert(isDir(dir));
  float theta = (3-dir) * (PI/2);
  rotMatrix(owner->transform,'Y',theta);
  owner->transform[3][0] = x;
  owner->transform[3][2] = z;
  owner->transform[3][1] = ty;
}

// place the car in the right place at an intersection - remember, we
// have to stay on the right side of the road
// also, we stop a few feet before the stop sign
void Drive::locateEntry(int it, int dir, float& x, float& z)
{
  // center of the intersection is a good place to start
  theIntersections(it)->getPosition(x,z);

  float dx,dz;
  direction(dir,dx,dz);

  // now move to the edge of the intersection, then 5 feet for the
  // stop sign (approx)
  x -= dx * (15 + 7);
  z -= dz * (15 + 7);

  // drive on the right side of the road
    x -= 7*dz;
    z += 7*dx;
}

void Drive::locateExit(int it, int dir, float& x, float& z)
{
  // center of the intersection is a good place to start
  theIntersections(it)->getPosition(x,z);

  float dx,dz;
  direction(dir,dx,dz);

  // right to the edge of the intersection
  x += dx * 15;
  z += dz * 15;

  // drive on the right side of the road
  x -= 7*dz;
  z += 7*dx;
}

void Drive::simulateUntil(unsigned long time)
{

  unsigned long t = time - stateStart;
  float alpha = ((float)t) / ((float) delay);
  if (alpha > 1) alpha = 1;
  float a1 = 1-alpha;

  //    printf("A:%4.3f s:%s i:%d d:%d n:%d nd:%d\n",
  //	 alpha,stateName(state),lastIntersect, dir, nextIntersect, newDir);

  switch(state) {
  case driving: {
	float nx = a1 * fromX + alpha * toX;
	float nz = a1 * fromZ + alpha * toZ;
	setMatrix(nx,nz,(float)dir);
	if (alpha >= .99999) {
	  state = stopped;
	  delay = 150 + rand() % 150;
	  lastIntersect = nextIntersect;
	  stateStart = time;
	  //printf("Stopping @ intersection %d (dir=%d)!\n",lastIntersect,dir);
	} }break;
  case intersection:
	stateStart = time;
	break;
  case stopped:
	// make sure we're at the stop sign - entry to an intersection
	reRange(dir);
	float x,z;
	locateEntry(lastIntersect,dir,x,z);
	setMatrix(x,z,dir);
	if ((time-stateStart) < delay) {
	} else {
	  // OK, time to figure out where to go next
	  nextIntersect = genNext(lastIntersect,dir, newDir);
	  assert((newDir >= 0) && (newDir <= 3));
	  getCurPos(fromX,fromZ);
	  if (newDir == dir) { // not turning, go straight through
		locateEntry(nextIntersect,dir,toX,toZ);
		state = driving;
		stateStart = time;
		delay = driveTime();
	  } else {
		stateStart = time;
		state = turning;
		locateExit(lastIntersect,newDir,toX,toZ);
		// rotate the right way around
		if ((newDir == 3) && (dir==0)) newDir=-1;
		if ((newDir == 0) && (dir==3)) newDir=4;
		delay = 300 + rand()%300;
	  }
	}
	break;
  case turning: {
	float nd = a1 * (float) dir + alpha * (float) newDir;
	float nx = a1 * fromX + alpha * toX;
	float nz = a1 * fromZ + alpha * toZ;
	setMatrix(nx,nz,nd);
	if (alpha >= .99999) {
	  stateStart = time;
	  state = driving;
	  dir = newDir % 4;
	  locateExit(lastIntersect,dir,fromX,fromZ);
	  locateEntry(nextIntersect,dir,toX,toZ);
	  delay = driveTime();
	}}
  break;
  };
  assert(isLegal(owner->transform[3][0]));
  assert(isLegal(owner->transform[3][1]));
  assert(isLegal(owner->transform[3][2]));
}

// how long till the next destination?
unsigned long Drive::driveTime()
{
  float d = ABS(fromX-toX) + ABS(fromZ-toZ);

  // that's in feet, now pick the speed
  float speed = 25 + ((float) (rand()%50));
  // now the number of seconds
  float t = d/speed;
  // convert to milliseconds...
  t *= 1000;
  return 50+t; 
}

void Drive::getCurPos(float& x, float& z)
{
	x = owner->transform[3][0];
	z = owner->transform[3][2];
}
